// kwmon.cc: update Jul 6 2004, LLB
// branched from kwmon.cc Jul 8 2004, LLB
// modified: Jul 13, 2004, LLB
// clustering added: Jul 26, 2004, LLB 
// chi-sq statistics added: July 28, 2004 LLB
// fixed chi-sq statistics (2A vs A degrees of freedom): Aug 12, 2004 LLB
// convert TSeries to floats in case of integer data (radio_lvea): Sep 15, 2004 LLB
// add default values for all options (lpefLength, filter, etc) to avoid crashes
//   if not defined: Sep 15, 2004 LLB
// add more precision to trigger output (code in header file): Sep 15, 2004 LLB
// add unnormalized energy output: Sep 20, 2004 LLB
// add version output and central time: Sep 22, 2004 LLB
// add support for software injections (2nd channel in same gwf file): Sep 27, 2004 LLB
// add some anticipatory flow control to prevent data loss at beginning of segment
//   in isolated cases: Nov 2, 2004 LLB
// changed lpefOrder to lpefLength and removed lpef::resetHistory due to changes in the LPEFilter class: Jan 16, 2005 SKC
// fixed call to logSignificance to account for factor of two: May 31, 2005 LLB
// add support for scaling of injection stream by a factor "factor": May 32, 2005 LLB
// add support for non-integer stride (samples per stride must be factor of two): May 32, 2005 LLB
// add support for multistream class (syntax for single stream dacc remains the same): June 29, 2005 LLB
// fix bug where nanosecond field is added twice to t0 (affects non-integer stride): July 13, 2005 LLB
// changed scale definitions to approximate frequency
// changed "mask" names to the more usual "segment"
// automatic linear filtering parameters: Aug 18, 2005 LLB
// add bound for linear filter order so that it does not go longer than 1 sec: Sep 14, 2005 LLB
//   still need to beware of ringing for low-frequency searches, should increase buffer time.
// move all functions into kleineWelle namespace: Sep 30, 2005 LLB
// add ability to include software injections from ASCII file: Jan 10, 2006 LLB
// extra comments and small changes: Jun 7, 2006 LLB
// increase transient duration and make it tunable: Jan 25, 2007 LLB

#include "mywv2lib.hh"
#include "kleineWelle.hh"
#include "gamma.cc"

#include <iterator>
#include <iostream>
#include <cstdlib>
#include <unistd.h>
#include <sstream>
#include <vector>
#include <cmath>

#define KWVERSION "Jan 25, 2007"

using namespace std;

EXECDAT(kleineWelle)

kleineWelle::kleineWelle(int argc, const char *argv[])
  : MultiStream(argc, argv), channel("H1:LSC-AS_Q"), injections(""), filter("lpef"),
    stride(16.0), significance(0.0), threshold(4.0), lpefLength(0), maxDist(2),
    lowFrequency(64), highFrequency(1024), lowScale(0), highScale(0), factor(1.0),
    decimateFactor(-1), transientDuration(4.0)
{
  bool syntax = false;  // flag for syntax error
  history = NULL;       // used for filtering in discontinuous data
  nextTime = Time(0);   // used to detect break in data stream
  segmentStart = 0;     // upper bound of segment
  fs = 0;               // sampling frequency to be defined at first data access
  wavePtr = NULL;
  prefilter = NULL; 
  lpefilter = NULL; 
//--------------------------------------- parse command syntax
  if(argc == 1) {
    syntax = true;
  }
  else {
    optionsFile.open(argv[1], ios::in);
    if(!(optionsFile.is_open())) {
      cerr << "cannot open options file: " << argv[1] << endl;
      finish(); return;
    }
  }
  if(syntax) {
    cerr << "kleineWelle version " << KWVERSION << endl;
    cerr << "Command syntax:" << endl;
    cerr << argv[0] << " [<options file>] [<MultiStream-args>]" << endl;
    cerr << "where the options file is a two column ascii file which defines the following options:";
    cerr << endl << "stride, transientDuration, channel, injections, factor, filter, threshold, maxDist, segmentFile, lowFrequency, highFrequency" << endl;
    cerr << "please see sample options file 'opt' for description of these parameters" << endl;
    finish(); return;
  }
  string option;
  string word;
//-------------------------------------- read in options from file
  while(!optionsFile.eof()) {
    word = "";
    getline(optionsFile, option);
    istringstream optionsStream(option);
    optionsStream >> word;
    if(word.length() == 0 || word[0] == '#')
      continue;
    else if(word == "stride") {         // DMT analysis stride
      double epoch;
      optionsStream >> epoch;
      stride = Interval(epoch);
    }
    else if(word == "channel")          // KW analysis channel
      optionsStream >> channel;
    else if(word == "injections")       // KW software injections channel
      optionsStream >> injections;
    else if(word == "factor")           // injections scale factor
      optionsStream >> factor;
    else if(word == "filter")           // prefilter type
      optionsStream >> filter;
    else if(word == "threshold")        // threshold on coefficient amplitude
      optionsStream >> threshold;
    else if(word == "significance")     // threshold on cluster significance
      optionsStream >> significance;
    else if(word == "maxDist")          // distance for clustering
      optionsStream >> maxDist;
    else if(word == "segmentFile") {    // segments to analyze
      string filename;
      optionsStream >> filename;
      segmentFile.open(filename.c_str(), ios::in);
      if(!(segmentFile.is_open())) {
        cerr << "cannot open segment file: " << filename << endl;
        finish(); return;
      }
    }
    else if(word == "waveFile") {        // ASCII file injection waveform
      string filename;
      optionsStream >> filename;
      waveFile.open(filename.c_str(), ios::in);
      if(!(waveFile.is_open())) {
        cerr << "cannot open waveform file: " << filename << endl;
        finish(); return;
      }
    }
    //------ most of these options are automatically calculated but take precedence if defined
    else if(word == "lpefLength")     // obsoleted (automatically decided)
      optionsStream >> lpefLength;
    else if(word == "lpefOrder")      // obsoleted
      optionsStream >> lpefLength;
    else if(word == "lowScale")       // obsoleted but take precedence if defined
      optionsStream >> lowScale;
    else if(word == "highScale")      // obsoleted but take precedence if defined
      optionsStream >> highScale;
    else if(word == "lowFrequency")   // low frequency of analysis (sets highScale)
      optionsStream >> lowFrequency;
    else if(word == "highFrequency")  // high frequency of analysis (sets lowScale)
      optionsStream >> highFrequency;
    else if(word == "decimateFactor") // number of times to decimate data stream (automatic)
      optionsStream >> decimateFactor;
    else if(word == "transientDuration") { // length of filter transient to account for
      double epoch;
      optionsStream >> epoch;
      transientDuration = Interval(epoch);
    }
    else
      cerr << "bad option: " << word << endl;
  }
  optionsFile.close();
  getDacc().setStride(stride);
  // set noise channel in stream 1
  // "sim" data needs "addSimulated", "adc" and "proc" data needs "addChannel"
  getDacc().addChannel(channel.c_str(), 0);
  // getDacc().addSimulated(channel.c_str(), 0);
  // set injections channel to stream 1 or 2 depending on number of streams
  if(!injections.empty()) {
    if(MultiStream::getNStream() <= 2) {
      getDacc().addChannel(injections.c_str(), MultiStream::getNStream() - 1);
      // getDacc().addSimulated(injections.c_str(), MultiStream::getNStream() - 1);
    }
    else {
      cerr << "kleineWelle currently supports <= 2 streams only" << endl;
      finish(); return;
    }
  }
  if(segmentFile.is_open()) {
    string segmentString;
    getline(segmentFile, segmentString);              // get the next segment, or EOF
    if(segmentFile.eof()) {
      finish();
      return;
    }
    istringstream segmentStream(segmentString);
    segmentStream >> segmentStart >> segmentStop;
    Time seektime(segmentStart, 0);
    cout << "# Analyzing segment " << segmentStart << " to " << segmentStop << endl;
    getDacc().seek(seektime);
  }
  cout << "# kleineWelle triggers" << endl;
  cout << "# Algorithm version: " KWVERSION << endl;
  cout << "# Analyzing channel: " << channel << endl;
  if(!injections.empty()) {
    cout << "# with injections: " << injections << endl;
    cout << "# with multiplication factor: " << factor << endl;
  }
  cout << "# with stride: " << stride << endl;
  cout << "# with transient duration: " << transientDuration << endl;
  cout << "# with black pixel threshold: " << threshold << endl;
  cout << "# with cluster signif threshold: " << significance << endl;
  cout << "# with maxDist between black pix: " << maxDist << endl;
}

kleineWelle::~kleineWelle()
{
  if(wavePtr) {
    delete   wavePtr;   wavePtr   = NULL;
  } 
  if(prefilter) {
    delete   prefilter; prefilter = NULL;
  }
  if(lpefilter) {
    delete   lpefilter; lpefilter = NULL;
  }
  if(segmentFile.is_open())
    segmentFile.close();
  cout << "# finished." << endl;
}

void kleineWelle::ProcessData(void) {
  Time t0;
  double* wavearray;  // array of wavelet coefficients
  int lsize;  // lsize keeps track of the start index of the wavelet coefficients in a scale
  TSeries* tsIn = getDacc().refData(channel.c_str());
  tsIn->Convert(4);        // Convert timeseries to floats (4) or doubles (5)
  if(!injections.empty()) {
    *(getDacc().refData(injections.c_str())) *= factor; // multiply injection stream by factor
    *tsIn += *(getDacc().refData(injections.c_str()));  // before adding it to noise time series
  }
  TSeries* tsOut = new TSeries();  // tsOut is a placeholder TSeries for the HP filtered data
  t0 = tsIn->getStartTime();
//------------------------------- set sample rate and filters if not defined
  if(!fs) {
    fs = int(tsIn->getNSample() / stride + 1e-6);
    printf("# detected sampling rate %d\n", fs);
    // need to cap high frequency boundary to nyquist
    if(highFrequency > fs / 2)
      highFrequency = fs /2;
    if(highScale)
      lowFrequency = fs / (1 << highScale);
    else
      highScale = (int)ceil(log2(fs / lowFrequency));
    if(lowScale)
      highFrequency = fs / (1 << lowScale);
    else {
      lowScale = (int)floor(log2(fs / highFrequency));
    }
    cout << "# over approximate frequency range " << lowFrequency << " - " << highFrequency << endl;
    cout << "# over wavelet scales " << lowScale << " to " << highScale << endl;
    if(decimateFactor == -1) {        // decimate factor is not set in options
      decimateFactor = lowScale - 1;  // keep some headroom for linear filter
    }
    cout << "# decimating factor " << decimateFactor << endl;
    FilterDesign fd(fs);
    fs = (fs >> decimateFactor);
    cout << "# decimating down to " << fs << endl;
    lowScale -= decimateFactor;
    highScale -= decimateFactor; 
    size = long(fs * double(stride) + 1e-6);
    wavePtr = new Haar_Wavelet(size);
    fd.decimateBy2(decimateFactor, 3);  // type 3 has most ringing, but least timeshift
    if(!filter.empty()) {  // with filtering defined add prefilters to fd filterdesign
      if(filter == "hpf" || filter == "lpef") {
	fd.butter(kHighPass, 6, lowFrequency, 0);
        cout << "# using high pass filtering at " << lowFrequency << endl;
	if(filter == "lpef" && !(lpefLength)) {
          // assume haar Q =~ 4, this provides sufficient whitening
          // actual form should be order > 'Q * fs / minFreq' as df = freq / Q
          // and df = fs / order
          // however do not exceed sample rate for filter order because we only have 1sec
          // to account for ringing (still not enough because of HPF ringing)
            lpefLength = min(fs, (int)(16 * highFrequency / lowFrequency));
	}
      }
      else if(filter != "none") {
        fd.filter(filter.c_str());  // add custom filter
        cout << "# using prefilter " << filter << endl;
      }
      else {
        cout << "# no prefilter" << endl;
      }
      if(lpefLength) {
	lpefilter = new LPEFilter(lpefLength, 0, (int)((double)stride * fs) - 1);
        cout << "# using linear predictive filtering of order " << lpefLength << endl;
      }
      // grab prefilter pipe, will contain decimator and possible hpf
      prefilter = fd.release();
    }
    cout << "# output: start, stop, central time, frequency, energy, normalized energy, "
	 << "number of pixels, significance" << endl;
    // read in waveform to inject from waveform file
    if(waveFile.is_open()) {
      string waveString;
      float sample;
      float hrss = 0;
      float hss = 0;
      int begin = size / 2;
      waveInjection = new float[size];
      for(int i = 0; i < size; i++) {
        waveInjection[i] = 0;
      }
      int length = 0;
      for(int i = 0; i < begin; i++) {  // fill second half of injection data with waveform
        getline(waveFile, waveString);
        if(waveFile.eof()) {
          break;
        }
        istringstream waveStream(waveString);
        waveStream >> sample;
        waveInjection[begin + i] = sample * factor;
        hss += sample * sample;
        length++;
      }
      hrss  = hss * factor / fs;
      cout << "# read in waveform injection of length " << length << " and scaled hrss of " << hrss << endl;
    }
  }

//------------- in the case of a break in data, bookeep for filter transients
  if(t0 != nextTime) {      // the current set of data is disontinuous from previous data
    printf("# detected break in data: expected %lu.%lu got %lu.%lu\n",
            nextTime.getS(), nextTime.getN(), t0.getS(), t0.getN());
    nextTime = t0 + getDacc().getStride();
    if(history != NULL) {   // if we just found a break in the data
                            // (this will only be 1 sec of data)
      delete history;       // kill the old history
      history = NULL;
      nextTime = Time(0);   // make sure we start from scratch next time
                            // (this second is lost)
      getDacc().setStride(stride); // reset the stride
    }
    else {
      // save all but the first second of data into the history to remove filter transients
      history = new TSeries(tsIn->extract(t0 + transientDuration,
                            stride - transientDuration));
      getDacc().setStride(transientDuration);   // next ProcessData we
                                            // only grab 1 second of data
      prefilter->reset();  // filter 1st second of data to store history
      *tsOut = prefilter->apply(tsIn->extract(t0, transientDuration));
      if(lpefLength) {
	lpefilter->setHistory(*tsOut);
      }
      delete tsOut;
      tsOut = NULL;
    }
  }
//------------------------ otherwise for continuous data, run the KW pipeline
  else {
    nextTime = t0 + getDacc().getStride();
    if(history != NULL) {  // last ProcessData we had found a break in the data
      history->Append(*tsIn);  // append the current second of data to the history
      *tsIn = *history;
      t0 = tsIn->getStartTime();
      delete history;
      history = NULL;
      getDacc().setStride(stride);  // reset stride back to normal
    }
    // for continuous data, add possible waveform injection
    if(waveFile.is_open()) {
      *tsIn += TSeries(tsIn->getStartTime(), tsIn->getTStep(), size, waveInjection);
      cout << "# added injection waveform vector" << endl;
    }
    *tsOut = prefilter->apply(*tsIn);
    // tsIn will hold the final filtered timeseries before the wavelet transform
    if(lpefLength) {
      lpefilter->train(*tsOut);
      lpefilter->apply(*tsOut, *tsIn);
    }
    else {
      *tsIn = *tsOut;
    }
  //-----------------------------------------------  do the wavelet transform
    tsIn->getData(size, wavePtr->getRawArray()); // put data into the wavelet class
    delete tsOut;
    tsOut = NULL;
    // test to see if we went off the segment bound, should rarely happen because of
    // anticipatory check at end of ProcessData (only happens for short segments < stride)
    if(segmentFile.is_open()) {
      if(t0 + stride > Time(segmentStop)) { // if we pass top bound of a line
	while(t0 + stride > Time(segmentStop)) {
          string segmentString;
          getline(segmentFile, segmentString);              // get the next segment, or EOF
          if(segmentFile.eof()) {
            finish();
            return;
          }
          istringstream segmentStream(segmentString);
          segmentStream >> segmentStart >> segmentStop;
          cout << "# Analyzing segment " << segmentStart << " to " << segmentStop << endl;
          if (t0 < Time(segmentStart)) {       // then skip to the next segment
            Time seektime(segmentStart, 0);
            getDacc().seek(seektime);
          }
        }
        return;
      }
    }
    printf("# processing at t0 = %lu.%lu\n", t0.getS(), t0.getN());
    wavePtr->transform(1);                       // do the transform
    vector<kleineWelle::Event> events;
    wavearray = wavePtr->getWaveArray(); // get the unnormalized wavelet coefficients
    float sigs[20] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    //-------------------------------------------  get statistics for wavelet coeffs
    for(int scale = lowScale; scale <= highScale; scale++) {
      lsize = size / (1 << scale);
      double lsum = 0;
      for(int i=lsize; i<2*lsize; i++) {
	lsum += wavearray[i]*wavearray[i];
	sigs[scale] = sqrt(lsum / lsize);
      }
    }
    //------------------------------- create vector of pixels which pass threshold
    vector<kleineWelle::Element> elements;
    for(int scale = lowScale; scale <= highScale; scale++) {
      lsize = size / (1 << scale);
      for(int i = lsize; i < 2*lsize; i++) {
        if(abs(wavearray[i] / sigs[scale]) > threshold) {
          double time = t0.getS() + 1e-9 * t0.getN();
          time += (double(1 << scale) / fs) * (i-lsize);
          elements.push_back(kleineWelle::Element(wavearray[i] * wavearray[i], wavearray[i] * wavearray[i] /
            (sigs[scale] * sigs[scale]), i, scale, time, time + (double(1 << scale) / fs),
            double(1 << scale) / fs, 0));
        }
      }
    }
    //------------------------------------ assign clusterID to each black pixel
    int freeID = 1;
    for(vector<kleineWelle::Element>::iterator i = elements.begin(); i < elements.end(); i++) {
      if(i->clusterID == 0) {
        i->clusterID = freeID;
        freeID = freeID + 1;
        kleineWelle::mergeClust(elements, i, maxDist);
      }
    }
    //------------------------------------- separate clusters by ID
    vector<kleineWelle::Element>* clusters;
    clusters = new vector<kleineWelle::Element>[freeID];
    for(vector<kleineWelle::Element>::iterator i = elements.begin(); i < elements.end(); i++) {
      clusters[i->clusterID].push_back(*i);
    }
    //--------------------------------------- estimate cluster parameters
    for(int i = 1; i < freeID; i++) {
      double startTime = 1e11;
      double endTime = 0;
      double centralTime = 0;
      double scale = 0;
      double energy = 0;
      double normEnergy = 0;
      int npts = 0;
      for(vector<kleineWelle::Element>::iterator j = clusters[i].begin(); j < clusters[i].end(); j++) {
        npts++;
        startTime = min(startTime, j->startTime);
        endTime = max(endTime, j->endTime);
        centralTime += j->normEnergy * (j->startTime + j->endTime) / 2.0;
        scale += j->scale * j->normEnergy;
        energy += j->energy;
        normEnergy += j->normEnergy;
      }
      scale = scale / normEnergy;
      centralTime = centralTime / normEnergy;
      double clusterSignificance = -gammqln(npts/2.0, normEnergy/2.0);
      double frequency = fs / pow(2, scale);
      if(clusterSignificance >= significance) { 
        events.push_back(kleineWelle::Event(startTime, endTime, centralTime, frequency, energy, normEnergy, npts,
                         clusterSignificance));
      }
    }
    delete[] clusters;
    clusters = NULL;
    //---------------------------------------- output sorted events
    sort(events.begin(), events.end());
    copy(events.begin(), events.end(), ostream_iterator<kleineWelle::Event>(cout));
    //--------------------- anticipate end of data stream as we approach segment end
    //                      the next ProcessData() will grab the next 'stride' of data
    if(segmentFile.is_open()) {
      if(t0 + 2.0 * stride > Time(segmentStop)) { // if the next data goes past masktop
        while(t0 + 2.0 * stride > Time(segmentStop)) {
          string segmentString;
          getline(segmentFile, segmentString);              // get the next segment, or EOF
          if(segmentFile.eof()) {
            finish();
            return;
          }
          istringstream segmentStream(segmentString);
          segmentStream >> segmentStart >> segmentStop;
          cout << "# Analyzing segment " << segmentStart << " to " << segmentStop << endl;
          if(t0+stride < Time(segmentStart)) { // then skip to the next segment
            Time seektime(segmentStart, 0);
            getDacc().seek(seektime);
          }
        }
        return;
      }
    }
  }
}

//--------------------------------------- recursive algorithm to assign clusterID
//                                        order N^2 not very smart!
void kleineWelle::mergeClust(vector<kleineWelle::Element>& elements, vector<kleineWelle::Element>::iterator i, int maxDist) {
  for(vector<kleineWelle::Element>::iterator j = elements.begin(); j < elements.end(); j++) {
    if(j->clusterID == 0) {
      if(kleineWelle::dist(*i, *j) <= maxDist) {
        j->clusterID = i->clusterID;
        kleineWelle::mergeClust(elements, j, maxDist);
      }
    }
  }
}

// negative natural logarithm of significance
double
kleineWelle::logSignificance(double normalizedEnergy, double timeFrequencyArea,
                double precision)
{

  /*
   * This function assumes time-frequency measurement of stationary white
   * Gaussian noise.  The resulting measurement (2*normalizedEnergy) are chi squared
   * distributed with 2A degrees of freedom, where A is the time-frequency
   * area of the measurement.  This function returns the negative natural
   * logarithm of the probability of achieving a less likely result.  It
   * uses a series expansion of the regularized upper incomplete gamma
   * function which is valid for the tail (highly significant) portion of
   * the distribution.  See Ambramowitz and Stegun, Handbook of Mathematical
   * Functions, section 6.1, page 264.
   */

  // maximum number of terms in series expansion
  int maximumTerms = 10;

  // initialize series expansion
  double sum = 1;
  double numerator = 1;
  double denominator = 1;
  double ratio;
  double term;

  // compute and sum series terms
  for (term = 1; term <= maximumTerms; term++) {  
    numerator *= (timeFrequencyArea - term) * (timeFrequencyArea > term);
    denominator *= normalizedEnergy;
    ratio = numerator / denominator;
    sum += ratio;
    if (ratio < precision * sum) {
      break;
    }
  }

  // give error if maximum number of terms was exceeded
  if (term > maximumTerms) {
    std::cerr << "logSignificance: maximum number of terms exceeded"
              << std::endl;
  }

  // negative natural logarithm of significance
  double logSignificance = normalizedEnergy -
                           (timeFrequencyArea - 1) * log(normalizedEnergy) -
                           log(sum) + lgamma(timeFrequencyArea);
  
  // return to calling function
  return logSignificance;
  
}
