/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef DACC_HH
#define DACC_HH

#include "DaccIn.hh"
#include "Time.hh"
#include "Interval.hh"
#include "Channel.hh"
#include <string>
#include <iosfwd>

class ChanList;   // Hide the channel list from root.
class TSeries;
class FSeries;

/**  Dacc methods are used to access individual data channels from frames.
  *  By decoupling the data access from the frame structure, Dacc relieves
  *  the Sandbox user from writing frame catination code, etc. The input 
  *  data are unpacked by Dacc into the TSeries objects.
  *  @memo LIGO Sandbox data access class.
  *  @author  John G. Zweizig
  *  @version 1.6; Last modified June 15, 2001
  */
class Dacc : public DaccIn
{
  public:
    //--------------------------------------  Constructors and destructors.
    /**  Default (Null) constructor.
      */
    Dacc(void);

    /**  Construct and connect to source.
      */
    explicit Dacc(const char* Source, const Time& start=Time(0));

#ifndef __CINT__
    /**  Construct using an already allocated frame reader.
      */
    explicit Dacc(FrameCPP::IFrameStream* reader);
#endif

    /**  Destructor.
      */
    ~Dacc(void);

    //---------------------------------------  Status manipulation
    /**  Set the DaccIn buffer count.
      *  @memo Set the DaccIn Buffer count.
      */
    void setBuffer(int N) {DaccIn::setBuffer(N);}

    /**  Set the DaccIn buffer count.
      *  @memo Set the DaccIn Buffer count.
      */
    void setChannelFlag(const std::string& chan, Channel::ChanFlags flag, 
			bool state=true);

    /**  If the debug flag is set, explanatory messages will be printed 
      *  when an error is encountered.
      *  @memo Set the debug flag.
      */
    void setDebug(int N) {DaccIn::setDebug(N);}

    /**  If the debug flag is set, explanatory messages will be printed 
      *  when an error is encountered.
      *  @memo Set the debug flag.
      */
    void setIgnoreMissingChannel(bool yn) {mIgnoreMissChan=yn;}

    /**  Set the noWait flag as specified. If noWait is set, the Dacc 
      *  methods that would otherwise wait for online data (e.g. synch, 
      *  fillData) will instead return with error code -8.
      *  @memo Set the noWait flag.
      */
    void setNoWait(bool now=true);

    /**  The default stride is used if a time interval is not specified
      *  when fillData() is invoked.
      *  @memo Define the default Stride.
      */
    void setStride(Interval Dt);

    //--------------------------------------  Channel list manipulation
    /**  The specified channel is added to the list of channels to be
      *  processed by #fillData#. #fillData# looks for the channel name 
      *  first in the FrAdcData list folowed by the FrProcData list. A 
      *  channel may be included in the list 
      *  only once. Subsequent calls to #addChannel# will replace the 
      *  original request. #TSptr# is used by the calling program to access 
      *  the data. If the pointer address is #NULL# or not specified, an 
      *  internal pointer is allocated, and the channel is accessed 
      *  exclusively with the #refData# method. The TSeries is then created 
      *  and destroyed under Dacc control. If the pointer address is 
      *  specified a non-zero pointer indicates a pre-allocated
      *  #TSeries#. If the pointer is #NULL# no #TSeries# is pre-allocated
      *  and the #TSeries# will be created when #fillData# is invoked. Note 
      *  that if the TSeries is not preallocated, the #TSeries# data vector 
      *  will have the same data type as the channel data (usually short). 
      *  #Decimate# gives the number of samples to be averaged together
      *  before they are entered into the TSeries. If #Decimate# <= 0, it
      *  is set to 1. If no TSeries is passed to #fillData# and the 
      *  #Decimation > 1#, a #float# data vector is allocated.
      *  @memo Add a channel to the request list.
      *  @param Name     Name of channel to collect data from.
      *  @param Decimate Decimation factor.
      *  @param TSptr    Address of TSeries pointer.
      */
    void addChannel(const char* Name, int Decimate=0, TSeries **TSptr=0);

    /**  Add an process frequency series to the list. The purpose of the 
      *  method is similar to that of #addChannel# except that the 
      *  channel being searched for is a processed data frequency series.
      *  @memo Add an FSeries data channel to the request list.
      *  @param Name  Name of channel to collect data from.
      *  @param FSptr Address of an FSeries pointer.
      */
    void addFSeries(const char* Name, FSeries **FSptr=0);

    /**  Add a Processed data channel to the list. The purpose of the method 
      *  and its arguments are identical to those for #addChannel# except 
      *  that any search for the channel is limited to Processed data.
      *  @memo Add a processed data channel to the request list.
      *  @param Name     Name of channel to collect data from.
      *  @param Decimate Decimation factor.
      *  @param TSptr    Address of TSeries pointer.
      */
    void addProcessed(const char* Name, int Decimate=0, TSeries **TSptr=0);

    /**  Add an Adc channel to the list. The purpose of the method and its
      *  arguments are identical to those for #addChannel# except that any 
      *  search for the channel is limited to the raw ADC data.
      *  @memo Add a raw data channel to the request list.
      *  @param Name     Name of channel to collect data from.
      *  @param Decimate Decimation factor.
      *  @param TSptr    Address of TSeries pointer.
      */
    void addRaw(const char* Name, int Decimate=0, TSeries **TSptr=0);

    /**  Add a Simulated data channel to the list. The purpose of the method 
      *  and its arguments are identical to those for #addChannel# except 
      *  that any search for the channel is limited to Simulated data.
      *  @memo Add a simulated data channel to the request list.
      *  @param Name     Name of channel to collect data from.
      *  @param Decimate Decimation factor.
      *  @param TSptr    Address of TSeries pointer.
      */
    void addSimulated(const char* Name, int Decimate=0, TSeries **TSptr=0);

    /**  The named channel is removed from the request list.
      *  @memo Remove a channel from the request list.
      */
    void rmChannel(const char* Name);

    /**  The data pointer for the frame currently in memory is incremented
      *  past the specified time. If the interval is not specified, 
      *  negative or zero, the rest of the data in the current frame are 
      *  skipped. If a discontinuity in the frame times is found, #flush#
      *  stops skipping at the start of the new data segment.
      *  @return Non-zero if an end of file is found.
      *  @memo Skip data.
      *  @param Stride Time interval to be skipped.
      */
    int flush(Interval Stride=Interval(0.0));

    /**  Data from channels specified by addChannel() are copied to TSeries
      *  objects.
      *  @memo   Fill the requested TSeries.
      *  @return 
      *  \begin{tabular}{|r|l|} \\
              0 & Successful completion \\
	     -1 & Frame start not contiguous to previous data \\
	     -2 & Sample rate isn't compatible with previous data. \\
	     -3 & Requested data not found in current frame \\
	     -4 & Error reading frame \\
	     -5 & Frame data are not self-consistent. \\
	     -6 & TSeries is not allocated. \\
	     -7 & Unsupported data type \\
	     -8 & Signal received while reading \\
	     -9 & Invalid data in structure.
      *  \end{tabular}
      *  @param Stride Time interval to be read in.
      *  @param start  Start a stride.
      */
    int fillData(Interval Stride=Interval(0.0), bool start=true);

    /**  Zero all TSeries in the channel list. Allocate sufficient
      *  memory to hold the given sample time.
      *  @memo   Reset channel TSeries.
      *  @param  Dt     Time interval to be used in allocating memory.
      */
    void zeroChans(Interval Dt);

    /**  Copy data to channel TSeries from a single frame.
      *  @memo   Fill the requested TSeries from a single frame.
      *  @return 
      *  \begin{tabular}{|r|l|} \\
              0 & Successful completion \\
	     -1 & Frame start not contiguous to previous data \\
	     -2 & Sample rate isn't compatible with previous data. \\
	     -3 & Requested data not found in current frame \\
	     -4 & Error reading frame \\
	     -5 & Frame data are not self-consistent. \\
	     -6 & TSeries is not allocated. \\
	     -7 & Unsupported data type \\
      *  \end{tabular}
      *  @param  Offset Offset of start of data from start of fram.
      *  @param  Dt     Time interval to be read in.
      */
    int fillChans(Interval Offset, Interval Dt);

    /**  Frames are read from the specified input file(s) until one 
      *  containing the specified time is found.
      *  @memo Skip to the specified time.
      *  @return Zero if open was successful.
      */
    int seek(Time STime=Time(0, 0));

    /**  The frame that will provide the data for the next read operation 
      *  is read into memory.
      *  @memo Synchronize the current frame buffer.
      *  @return Zero if successful or -1 if unable to read in a frame.
      */
    int synch(void);

    //---------------------------------  Accessors
    /**  Returns the current data collection time i.e. the time of the 
      *  first sample not used by fillData. getCurrentTime() returns 
      *  Time(0) if no frame has been read. Note that data filling may 
      *  not start at getCurrentTime if there are gaps in the data stream
      *  or if filData errors occur. Use synch to position the current 
      *  time to the start of the next frame if current frame is not 
      *  available.
      *  @memo Get the Current time.
      *  @return Start time for next data fill.
      */
    Time getCurrentTime(void) const {return getTime()+mOffset;}

    /**  Returns the level of debug information that was requested.
      *  @memo Get the debug flag.
      *  @return Debug print level.
      */
    int getDebug(void) {return DaccIn::getDebug();}

    /**  Returns the time at which data collection started with the
      *  last invocation of fillData.
      *  @memo Get the start time of the last fill.
      *  @return Start time of last fillData stride.
      */
    Time getFillTime(void) const {return mFillTime;}

    /**  Return a pointer to message text for the specified error.
      *  @memo Get error message text.
      *  @return Constant pinter to text.
      */
    static const char* getMsgText(int err);

    /**  Returns an offset of the current point in the frame.
      *  @memo Get the current offset into the frame.
      */
    Interval getOffset(void) const {return mOffset;}

    /**  Returns the time interval over which data are collected.
      *  @memo Get the stride interval.
      */
    Interval getStride(void) const {return mStride;}

    /**  Test to see is the names channel has been requested..
      *  @memo Test if channel was requested.
      *  @return True if the specified channel has been requested.
      *  @param Name Channel name to be tested.
      */
    bool isChannelRead(const char* Name) const;

    /**  A formatted list of requested channels, their decimation factors, 
      *  and latest time copied is printed to the specified output stream. 
      *  @memo List requested channels.
      */
    std::ostream& list(std::ostream& out) const;

    /**  Returns a pointer to the last TSeries filled for the named channel.
      *  @memo Get a pointer to the data.
      */
    const TSeries* refData(const char* name) const;
    TSeries* refData(const char* name);

    /**  Returns a pointer to the last TSeries filled for the named channel.
      *  @memo Get a pointer to the data.
      */
    const FSeries* refFData(const char* name) const;
    FSeries* refFData(const char* name);

  private:
    typedef std::list<Channel> ChannelList;
    typedef ChannelList::iterator chan_iter;
    typedef ChannelList::const_iterator const_chan_iter;

    const_chan_iter findChannel(const std::string& Name, 
				Channel::ChanType t=Channel::kUnknown) const;
    chan_iter       findChannel(const std::string& Name, 
				Channel::ChanType t=Channel::kUnknown);

  private:
    /**  Start time of last fill.
      *  @memo Time of last fill.
      */
    Time mFillTime;

    /**  Time offset of the first unused sample within the current frame.
      *  @memo Time offset of the first unused sample in the current frame.
      */
    Interval mOffset;

    /**  Default data collection time interval (stride).
      *  @memo Default data collection time interval (stride).
      */
    Interval mStride;

    /**  List of requested channels.
      *  @memo List of requested channels.
      */
    ChannelList mChanList;

    /**  Flag indicating whether a fill is in progress. mFillRun is set
      *  to true by fillData() at the start of a fill. It remains true
      *  if the fill is interrupted by a signal while waiting for data.
      *  It is reset to false when the requested time series are filled
      *  with data, or when an unrecoverable error occurs.
      */
    bool     mFillRun;

    /**  Number of seconds of data already copied into the TSeries by the 
      *  current (last) fillData run.
      */
    Interval mFillOffset;

    /** Stride used by the current (last fill).
     */
    Interval mFillStride;

    /** Don't wait for data if none is currently available.
      */
    bool     mNoWait;

    /** If set, missing channels are ignored while filling data.
      */
    bool     mIgnoreMissChan;
};
#endif // DACC_HH
